import os
import abc

from .models import Message


class LLM(abc.ABC):
    async def chat(self, messages: list[Message], **kwargs):
        result = []

        async for chunk in self.chat_stream(messages, **kwargs):
            result.append(chunk)

        return "".join(result)

    @abc.abstractmethod
    async def chat_stream(self, messages: list[Message], **kwargs):
        pass

    @abc.abstractmethod
    async def complete_stream(self, prompt: str, **kwargs):
        pass


try:
    from openai import AsyncOpenAI
    from openai.types.chat import ChatCompletionChunk
    from openai.types.completion import Completion

    class OpenAILLM(LLM):
        def __init__(self, model: str, api_key: str, base_url: str = None) -> None:
            self.client = AsyncOpenAI(api_key=api_key, base_url=base_url)
            self.model = model

        async def chat_stream(self, messages: list[Message], **kwargs):
            stream = await self.client.chat.completions.create(
                messages=[dict(role=m.role, content=m.content) for m in messages],
                model=self.model,
                stream=True,
                **kwargs,
            )

            async for response in stream:
                r: ChatCompletionChunk = response
                yield r.choices[0].delta.content or ""

        async def complete_stream(self, prompt: str, **kwargs):
            stream = await self.client.completions.create(
                model=self.model, prompt=prompt, stream=True, **kwargs
            )

            async for chunk in stream:
                r: Completion = chunk
                yield r.choices[0].text or ""

except ImportError:
    pass



import subprocess
import importlib


class Tool:
    skip_use = False

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def description(self) -> str:
        return self.__class__.__doc__

    def describe(self) -> str:
        return f"- {self.name}: {self.description}."

    def prompt(self, query) -> str:
        return query

    def use(self, query, response):
        pass

    def conclude(self, query, output):
        pass


class Chat(Tool):
    """
    When the user engages in general-purpose or casual conversation.
    """

    def __init__(self) -> None:
        self.skip_use = True


class Bash(Tool):
    """
    When the user requests some action in the filesystem or terminal,
    including git commands, or installing new applications or packages.
    """

    def prompt(self, query) -> str:
        return f"""
Given the following user query, generate a single bash command line
that performs the indicated functionality.

Reply only with the corresponding bash line.
Do not add any explanation.

Query: {query}
Command:
"""

    def use(self, query, response):
        response = response.strip("`")

        if response.startswith("bash"):
            response = response[4:]

        response = response.split("`")[0]
        response = [s.strip() for s in response.split("\n")]
        response = [s for s in response if s]

        response = ";".join(s for s in response)

        yield "Running the following code:\n"
        yield "$ "
        yield response
        yes = input("\n[y]es / [N]o ")

        if yes != "y":
            yield "(!) Operation cancelled by your request.\n"
            return

        yield "\n"

        p = subprocess.run(response, shell=True, stdout=subprocess.PIPE)
        yield p.stdout.decode("utf8")

    def conclude(self, query, output):
        return f"""
The user issued the following query:

Query: {query}

Given that query, you ran the following command which
Add license information here.
produced the given output:

---
{output}
---

If the user query is a question, answer it as succintly
as possible given the output.

If the user query was a request to do something,
explain briefly the result of the operation.
"""


class Interpreter(Tool):
    """
    When the user asks a mathematical question that can
    be solved with a simple Python function.
    """

    def prompt(self, query) -> str:
        return f"""
Given the following user query,
generate a single Python function named `solve`
and the necessary import statements
to perform the indicated functionality.

If you need secondary functions, name them starting with `_`.

Enclose the code in ```python and ```

Reply only with the corresponding Python code.
Do not add any explanation.
Do not execute the function.
Do not add any print statements.

Query: {query}
Function definition:
"""

    def use(self, query, response):
        code = []
        imports = []
        inside = False

        for line in response.split("\n"):
            if line.startswith("```python"):
                inside = True
            elif line.startswith("```"):
                inside = False
            elif inside:
                if line.startswith("import"):
                    imports.append(line.split()[1])
                else:
                    code.append(line)

        code.append("\nresult = solve()")
        code = "\n".join(code).strip()

        yield "Will run the following code:\n\n"
        yield code
        yes = input("\n\n[y]es / [N]o ")

        if yes != "y":
            yield "(!) Operation cancelled by your request.\n"
            return

        globals = {module: importlib.import_module(module) for module in imports}
        locals = {}
        exec(code, globals, locals)
        result = locals["result"]

        yield f"\nResult: {result}"


class Codegen(Tool):
    """
    When the user makes a general question about programming or
    explicitly asks to generate code in a given programming language.
    """

    def __init__(self) -> None:
        self.skip_use = True

    def prompt(self, query) -> str:
        return f"""
Answer the following user query about programming with
a general explanation in broad terms, followed by
one or more examples of code, as necessary.

Enclose all code examples in ``` with the corresponding
programming language identifier.

Query: {query}
"""



import os
import dotenv
import asyncio
import argparse
from .core import Agent
from .connectors import LLM, OpenAILLM
from .tools import Bash, Chat, Codegen, Interpreter


def run():
    dotenv.load_dotenv()

    parser = argparse.ArgumentParser("lovelaice")
    parser.add_argument(
        "-f", "--file", action="store", help="Add a file to the context"
    )
    parser.add_argument(
        "-m",
        "--model",
        action="store",
        help="Model to use",
        default=os.getenv("LOVELAICE_MODEL"),
    )
    parser.add_argument(
        "--base-url",
        action="store",
        help="API base URL to use",
        default=os.getenv("LOVELAICE_BASE_URL"),
    )
    parser.add_argument(
        "--api-key",
        action="store",
        help="API key to use",
        default=os.getenv("LOVELAICE_API_KEY"),
    )
    parser.add_argument(
        "-c",
        "--complete",
        action="store_true",
        help="Instead of full chat interaction, simply run completion on the input prompt.",
        default=False,
    )
    parser.add_argument(
        "--max-tokens",
        action="store",
        type=int,
        help="Max number of tokens allowed to generate. Defaults to 2048.",
        default=int(os.getenv("LOVELAICE_MAX_TOKENS", 2048)),
    )
    parser.add_argument(
        "--min-words",
        action="store",
        type=int,
        help="Only for completion mode, keep running completion until at least this number of words are generated.",
        default=0,
    )
    parser.add_argument("query", nargs="*", default=None)

    args = parser.parse_args()

    llm = OpenAILLM(args.model, args.api_key, args.base_url)

    if args.complete:
        asyncio.run(complete(args, llm))
        return

    agent = Agent(
        llm,
        tools=[Bash(), Chat(), Interpreter(), Codegen()],
    )

    if args.query:
        asyncio.run(run_once(args, agent))
    else:
        asyncio.run(run_forever(args, agent))


async def complete(args, llm: LLM):
    prompt = " ".join(args.query)

    print(prompt, end="", flush=True)

    while True:
        generated = False

        async for chunk in llm.complete_stream(prompt, max_tokens=args.max_tokens):
            prompt += chunk
            print(chunk, end="", flush=True)

            if chunk:
                generated = True

        if not generated or len(prompt.split()) > args.min_words:
            break

    print()


async def run_once(args, agent: Agent):
    prompt = " ".join(args.query)

    async for response in agent.query(prompt, max_tokens=args.max_tokens):
        print(response, end="", flush=True)

    print()


async def run_forever(args, agent: Agent):
    while True:
        try:
            prompt = input("> ")

            try:
                async for response in agent.query(prompt, max_tokens=args.max_tokens):
                    print(response, end="", flush=True)
            except asyncio.exceptions.CancelledError:
                print("(!) Cancelled")

            print("\n")
        except KeyboardInterrupt:
            break
        except EOFError:
            break



from .connectors import LLM
from .models import Message
from .tools import Chat, Tool


SYSTEM_PROMPT = """
You are Lovelaice, a helpful AI agent that runs in the
user terminal. You can do many things, from casual chat
to helping with concrete tasks that require access to the user filesystem.
You can run programs at the request of the user, and write
code for them.
You have access to the Internet and can answer queries that
require searching online.
"""

TOOLS_PROMPT = """
These are some of the tools you have access to:

{tools}

According to the following query, reply only with the name of the tool that
is most appropriate to solve the problem. Do not reply anything else.

Query: {query}
Tool:
"""


class Agent:
    def __init__(self, client: LLM, tools: list[Tool]) -> None:
        self.client = client
        self.tools = tools
        self.tools_dir = {t.name: t for t in tools}
        self.tools_line = "\n".join(t.describe() for t in tools)

    async def query(
        self,
        prompt: str,
        use_tool=None,
        **kwargs,
    ):
        if use_tool is None:
            messages = [
                Message(role="system", content=SYSTEM_PROMPT),
                Message(
                    role="user",
                    content=TOOLS_PROMPT.format(query=prompt, tools=self.tools_line),
                ),
            ]

            tool_name = await self.client.chat(messages, **kwargs)
            tool_name = tool_name.split()[0].strip(",.:")

            async for response in self.query(prompt, use_tool=tool_name, **kwargs):
                yield response

        else:
            if use_tool not in self.tools_dir:
                tool: Tool = Chat()
            else:
                tool: Tool = self.tools_dir[use_tool]

            if tool.name != "Chat":
                yield f":: Using {tool.name}\n\n"

            messages = [Message(role="user", content=tool.prompt(prompt))]

            if tool.skip_use:
                async for response in self.client.chat_stream(messages, **kwargs):
                    yield response

            else:
                response = await self.client.chat(messages, **kwargs)
                output = []

                for line in tool.use(prompt, response):
                    output.append(line)
                    yield line

                output = "\n".join(output)
                conclusion = tool.conclude(prompt, output)

                if conclusion is None:
                    return

                messages = [Message(role="user", content=conclusion)]

                yield "\n"

                async for response in self.client.chat_stream(messages, **kwargs):
                    yield response



from lovelaice.models import Parser


def test_title():
    sentences = ["Title: a new hope", "lorem ipsum dolor sit amet"]
    parser = Parser()

    assert parser.parse(sentences) == ["# A new hope", "lorem ipsum dolor sit amet"]


def test_heading():
    sentences = ["Heading, a new hope", "lorem ipsum dolor sit amet"]
    parser = Parser()

    assert parser.parse(sentences) == ["## A new hope", "lorem ipsum dolor sit amet"]


def test_subheading():
    sentences = ["Subheading a new hope", "lorem ipsum dolor sit amet"]
    parser = Parser()

    assert parser.parse(sentences) == ["### A new hope", "lorem ipsum dolor sit amet"]


def test_ul():
    sentences = ["Something", "Begin unordered list.", "Item 1.", "Item 2.", "end unordered list", "Something else"]
    parser = Parser()

    assert parser.parse(sentences) == [
        "Something",
        "- Item 1.",
        "- Item 2.",
        "Something else",
    ]


def test_ol():
    sentences = ["Something", "Begin ordered list.", "Item 1.", "Item 2.", "end ordered list", "Something else"]
    parser = Parser()

    assert parser.parse(sentences) == [
        "Something",
        "1. Item 1.",
        "2. Item 2.",
        "Something else",
    ]


def test_tl():
    sentences = ["Something", "Begin todo list.", "Item 1.", "Item 2.", "end todo list", "Something else"]
    parser = Parser()

    assert parser.parse(sentences) == [
        "Something",
        "- [ ] Item 1.",
        "- [ ] Item 2.",
        "Something else",
    ]
